/*
// File:		EmaiMessagePartEnumerator.cpp
//
// Version:		mail.it library 1.0 sample file
//
// Copyright:	(C) 2006 Dacons LLP.
//
// Bugs:		If you have problems or suggestions, please contact
//				the Online Support at: http://www.dacons.net/support
*/

#include PCH_FILE_NAME
#include "MessagePartEnumerator.hpp"
#include "StreamUtils.hpp"

static EmaiUniChar kFileName[]		= {'f', 'i', 'l', 'e', 'n', 'a', 'm', 'e', 0};
static EmaiUniChar kName[]			= {'n', 'a', 'm', 'e', 0};

const EmaiUint32 kCounterKey = 'COUN';

Emai::ContactList&
MessagePartEnumerator::GetContactStorage(
		EmaiOptions inOptions
	)
{
	/*Determine the type of the contact field*/
	switch (inOptions)
	{
		case EmaiFrom:
			return mContactFrom;
			break;

		case EmaiTo:
			return mContactTo;
			break;

		case EmaiSender:
			return mContactSender;
			break;

		case EmaiReplyTo:
			return mContactReplyTo;
			break;

		case EmaiCC:
			return mContactCC;
			break;

		case EmaiBCC:
			return mContactBCC;
			break;

		default:
			break;
	}
	assert(false);
	Emai::CheckErrorCode(EmaiErrIncorrectParameter);
	return mContactFrom;
}

/*Handle the contact fields receiving*/
/*Parameters:*/
/*	inContacts - pointer to the array of the EmaiContact elements*/
/*	inContactsCount - size of the array of the EmaiContact elements*/
/*	inOptions - type of the contact field*/
void
MessagePartEnumerator::OnContacts(
		Emai::ContactList inContactList,
		EmaiOptions inOptions
	)
{
	/*contacts are succesfully retrieved*/
	/*copy them in local storage*/
	Emai::ContactList& contactList = GetContactStorage(inOptions);

	contactList = inContactList;
}


/*Handle the message subject field receiving*/
/*Parameters:*/
/*	inSubject - pointer to the subject data of the messaage*/
void
MessagePartEnumerator::OnSubject(
		EmaiConstUniCharPtr inSubject
	)
{
	/*Save the pointer to the subject data*/
	mSubject = inSubject;
}

/*Handle the message plain text part receiving*/
/*	inPart - reference to the message text part*/
void
MessagePartEnumerator::OnPlainTextPart(
		Emai::MessagePart& inPart,
		const EmaiPartData& /*inPartData*/
	)
{
	/*Create the message text part*/
	mTextParts.push_back(inPart);
}

/*Handle the message unicode text part receiving*/
/*Paramters:*/
/*	inPart - reference to the message unicode text part*/
void
MessagePartEnumerator::OnUnicodePlainTextPart(
		Emai::MessagePart& inPart,
		const EmaiPartData& /*inPartData*/
	)
{
	/*Create the message unicode text part*/
	mTextParts.push_back(inPart);
}

/*Handle the message html text part receiving*/
/*Parameters:*/
/*	inPart - reference to the meesage html part*/
void
MessagePartEnumerator::OnHtmlTextPart(
		Emai::MessagePart& inPart,
		const EmaiPartData& /*inPartData*/
	)
{
	//Store message html part
	mHtmlPart = inPart;
}

/*Handle the message data receiving event*/
/*Parameters:*/
/*	inPart - reference to the message data*/
/*	inPartData - reference to the message part data*/
void
MessagePartEnumerator::OnDataPart(
		Emai::MessagePart& inPart,
		const EmaiPartData& inPartData
	)
{
	//Check the presence of the message data identifier
	if (inPartData.contentID == NULL)
	{
		//Create the attachment and add it to the list of the attachments
		mAttachments.push_back(inPart);
	}
	else
	{
		//Create the inline attachment and add it to the list of the inline attachments
		mInlines.push_back(inPart);
	}

	EmaiConstUniCharPtr filename = GetFileName(inPartData);
	if (filename == NULL)
	{
		//The part doesn't have filename specified, we should generate one later
		//Increase internal counter and save it as user data in the message part
		EmaiUint32 counter = 0;
		EmaiUint32 size = sizeof(counter);
		mMessage.GetUserData(kCounterKey, &counter, size);
		++counter;
		mMessage.SetUserData(kCounterKey, &counter, sizeof(counter));
		inPart.SetUserData(kCounterKey, &counter, sizeof(counter));
	}
}

/*Handle the composite message part receive beginning*/
/*Parameters:*/
/*	inPartData - reference to the message part data*/
void
MessagePartEnumerator::OnCompositePartStart(
		Emai::MessagePart& /*inPart*/,
		const EmaiPartData& inPartData
	)
{
	OnPart(inPartData);
}

/*Handle the composite message part receive ending*/
/*Parameters:*/
/*	inPartData - reference to the message part data*/
void
MessagePartEnumerator::OnCompositePartEnd(
		Emai::MessagePart& /*inPart*/,
		const EmaiPartData& inPartData
	)
{
	OnPart(inPartData);
}

/*Handle the message part receiving event*/
void
MessagePartEnumerator::OnPart(
		const EmaiPartData& /*inPartData*/
	)
{
}

void
MessagePartEnumerator::PreProcessMessageParts()
{
	EmaiUint32 counter = 0;
	mMessage.SetUserData(kCounterKey, &counter, sizeof(counter));
}

void
MessagePartEnumerator::PostProcessMessageParts()
{
	mMessage.SetUserData(kCounterKey, NULL, 0);
}

/*Return the pointer to the message subject*/
const Emai::UniString&
MessagePartEnumerator::GetSubject(
	)
{
	return mSubject;
}

/*Return list of the message contacts*/
/*	inOptions - type of the message contact*/
Emai::ContactList&
MessagePartEnumerator::GetContacts(
		EmaiOptions inOptions
	)
{
	return GetContactStorage(inOptions);
}

/*Return the message html data and its length*/
/*Parameters:*/
/*	outText - on output contains message html text*/
void
MessagePartEnumerator::GetHtml(
		Emai::String& outHtml
	)
{
	outHtml.clear();
	if (mHtmlPart != NULL)
	{
		EmaiPartData outData;
		
		mHtmlPart.GetData(outData);

		if (outData.partData.htmlText.html != NULL)
			outHtml.assign(outData.partData.htmlText.html, outData.partData.htmlText.htmlSize);
	}
}

/*Return the number of elements in the attachments list*/
EmaiUint32
MessagePartEnumerator::GetAttachmentsCount()
{
	return static_cast<EmaiUint32>(mAttachments.size());
}

/*Return the message attachment data and its length*/
/*Parameters:*/
/*	inBuffer - pointer to the message attachment data*/
/*	outLength - length of the message attachment data*/
void
MessagePartEnumerator::GetAttachmentAtIndex(
		const EmaiUint32 inIndex,
		Emai::UniString& outFileName,
		EmaiCharPtr outBuffer,
		EmaiUint32& ioLength
	)
{
	outFileName.clear();
	EmaiPartData data;
		
	mAttachments[inIndex].GetData(data);

	if (outBuffer == NULL)
	{
		/*Return only the length of the message attachment data, in case if*/
		/*the message attachment data pointer is NULL*/
		ioLength = data.partData.data.dataSize;
	}
	else
	{
		/*Returns the message attachment data and its length*/
		ioLength = (ioLength < data.partData.data.dataSize) ? ioLength : data.partData.data.dataSize;
		memcpy(outBuffer, data.partData.data.data, ioLength);
	}

	EmaiConstUniCharPtr filename = GetFileName(data);
	if (filename == NULL)
		GenerateFileName(mAttachments[inIndex], outFileName);
	else
		outFileName = filename;
}

/*Return the number of elements in the inline attachments list*/
EmaiUint32
MessagePartEnumerator::GetInlinesCount()
{
	return static_cast<EmaiUint32>(mInlines.size());
}


/*Returns the message inline attachment data and its length*/
/*Parameters:*/
/*	outBuffer - pointer to the message inline attachment data*/
/*	outLength - length of the message inline attachment data*/
void
MessagePartEnumerator::GetInlineAtIndex(
		const EmaiUint32 inIndex,
		Emai::UniString& outFileName,
		Emai::UniString& outContentID,
		EmaiCharPtr outBuffer,
		EmaiUint32& ioLength
	)
{
	outFileName.clear();
	outContentID.clear();
	EmaiPartData data = {0};
		
	mInlines[inIndex].GetData(data);
	
	if (outBuffer == NULL)
	{
		/*Return only the length of the message inline attachment data, in case if*/
		/*the message inline attachment data pointer is NULL*/
		ioLength = data.partData.data.dataSize;
	}
	else
	{
		/*Returns the message inline attachment data and its length*/
		ioLength = (ioLength < data.partData.data.dataSize) ? ioLength: data.partData.data.dataSize;

		memcpy(outBuffer, data.partData.data.data, ioLength);
	}
	
	if (data.contentID != NULL)
		outContentID = data.contentID;
	
	EmaiConstUniCharPtr filename = GetFileName(data);
	if (filename == NULL)
		GenerateFileName(mInlines[inIndex], outFileName);
	else
		outFileName = filename;
}


/*Return the message plain text data and its length*/
/*Parameters:*/
/*	outText - on output contains message plain text*/
void
MessagePartEnumerator::GetText(
		Emai::String& outText
	)
{
	outText.clear();
	EmaiPartData outData;
	for (EmaiUint32 i = 0; i < mTextParts.size(); i++)
	{
		mTextParts[i].GetData(outData, EmaiGetNativeTextData);	
		if (outData.partData.plainText.text != NULL)
			outText.append(outData.partData.plainText.text, outData.partData.plainText.textSize);
	}
}

/*Return the message unicode text data and its length*/
/*Parameters:*/
/*	outText - on output contains message unicode plain text*/
void
MessagePartEnumerator::GetUnicodeText(
		Emai::UniString& outText
	)
{
	outText.clear();
	EmaiPartData outData;
	for (EmaiUint32 i = 0; i < mTextParts.size(); i++)
	{
		mTextParts[i].GetData(outData);
		if (outData.partData.unicodeText.text != NULL)
			outText.append(outData.partData.unicodeText.text, outData.partData.unicodeText.textSize);
	}
}

/*Return the number of elements in the text parts list*/
EmaiUint32
MessagePartEnumerator::GetTextPartsCount()
{
	return static_cast<EmaiUint32>(mTextParts.size());
}

/*Return the message unicode text data and its length*/
/*Parameters:*/
/*	inIndex - index of the text part to be returned*/
/*  outText - string to store text*/
void 
MessagePartEnumerator::GetUnicodeTextAtIndex(
		const EmaiUint32 inIndex, 
		Emai::UniString& outText
	)
{
	outText.clear();
	EmaiPartData outData;
	mTextParts[inIndex].GetData(outData);
	
	if (outData.partData.unicodeText.text != NULL)
		outText.assign(outData.partData.unicodeText.text, outData.partData.unicodeText.textSize);
}

/*Return the message plain text data and its length*/
/*Parameters:*/
/*	inIndex - index of the text part to be returned*/
/*  outText - string to store text*/
void 
MessagePartEnumerator::GetTextAtIndex(
		const EmaiUint32 inIndex,
		Emai::String& outText
	)
{
	outText.clear();
	EmaiPartData outData;
	mTextParts[inIndex].GetData(outData, EmaiGetNativeTextData);

	if (outData.partData.plainText.text != NULL)
		outText.assign(outData.partData.plainText.text, outData.partData.plainText.textSize);
}

/*Saves attachment to  the file system*/
/*Parameters:*/
/*	inIsInline - chose inline or ordinary attachment*/
/*	inIndex - index of the attachment to be saved*/
/*  inPath - file system path for saving*/
bool 
MessagePartEnumerator::SaveAttachmentAtIndex(
		const bool inIsInline,
		const EmaiUint32 inIndex,
		EmaiConstUniCharPtr inPath,
		EmaiConstUniCharPtr inNewName
	)
{
	EmaiPartData data;

	if (inIsInline)
		mInlines[inIndex].GetData(data);
	else 
		mAttachments[inIndex].GetData(data);

	FILE* file = NULL;
	Emai::UniString filePath = inPath;
	AppendPathDelimiter(filePath);

	Emai::UniString resFileName;
	if (inNewName == NULL)
	{
		EmaiConstUniCharPtr filename = GetFileName(data);
		if (filename == NULL)
		{
			if (inIsInline)
				GenerateFileName(mInlines[inIndex], resFileName);
			else 
				GenerateFileName(mAttachments[inIndex], resFileName);
		}
		else
			resFileName = filename;
	}
	else
		resFileName = inNewName;

	filePath += resFileName;
	
	file = OpenAttachmentFile(filePath);
	if (file == NULL)
		return false;
	::fwrite(data.partData.data.data, data.partData.data.dataSize, 1, file);
	::fclose(file);

	return true;
}

/*Return attachment file name for current part data*/
/*Parameters:*/
/*	inPartData - part data struct*/
EmaiConstUniCharPtr
MessagePartEnumerator::GetFileName(
		const EmaiPartData& inPartData
	)
{
	/*Get the pointer to file name of the message attachment*/
	EmaiConstUniCharPtr filename = Emai::ContentField(
			inPartData.contentDisposition, true).FindParameterValue(kFileName);
	if (filename == NULL)
		filename = Emai::ContentField(inPartData.contentType,
									true).FindParameterValue(kName);
	return filename;
}

void
MessagePartEnumerator::GenerateFileName(
		Emai::MessagePart& inPart,
		Emai::UniString& outFileName
	)
{
	EmaiChar buff[256] = "";
	EmaiUint32 counter = 0;
	EmaiUint32 size = sizeof(counter);
	inPart.GetUserData(kCounterKey, &counter, size);
	sprintf(buff, "ATT%d", counter);

	EmaiConstCharPtr curr = buff;
	while (curr[0] != '\0')
	{
		outFileName += curr[0];
		++curr;
	}
}

/*Return string with backslashe*/
/*Parameters:*/
/*	ioPath - string with or without backslash*/
void
MessagePartEnumerator::AppendPathDelimiter(
		Emai::UniString& ioPath
	)
{
#ifdef __GNUC__
	EmaiUniChar slash[] = {'/', 0};
#else
	EmaiUniChar slash[] = {'\\', 0};
#endif
	if (ioPath[ioPath.length() - 1] != slash[0])
		ioPath += slash;
}

/*Convert wide-character string to a new character string. Return character string*/
/*Parameters:*/
/*	inStr - UTF-16 string*/
FILE*
MessagePartEnumerator::OpenAttachmentFile(
		const Emai::UniString& inFilePath
	)
{
	FILE* file = NULL;
#ifdef __GNUC__
	CFStringRef str = ::CFStringCreateWithCharactersNoCopy(kCFAllocatorDefault,
						inFilePath.c_str(), inFilePath.length(), kCFAllocatorNull);
	assert(0 != str);
	CFURLRef url = ::CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
							str, kCFURLPOSIXPathStyle, false);
	assert(0 != url);
	EmaiChar path[PATH_MAX];
	bool bSuccess = ::CFURLGetFileSystemRepresentation(url, false,
						reinterpret_cast<UInt8*>(path), sizeof(path));
	assert(bSuccess);
	::CFRelease(url);
	::CFRelease(str);
	file = ::fopen(path, "w+b");
#else
	file = ::_wfopen(inFilePath.c_str(), L"w+b");
#endif
	return file;
}
